"""Utilities for tests"""

import copy
import re

BAD_ID = "line %s: id '%s' doesn't match '%s'"
BAD_SEQLEN = "line %s: %s is not the same length as the first read (%s)"
BAD_BASES = "line %s:  %s is not in allowed set of bases %s"
BAD_PLUS = "line %s: expected '+', got %s"
BAD_QUALS = "line %s: %s is not the same length as the first read (%s)"
MSG_INCOMPLETE = "incomplete record at end of file %s"


class Fastq:
    """A convenient data structure for handling the fastqs generated by qasim.
    NOTES:
    * Read id's are the form: @NAME_COORD1_COORD2_ERR1_ERR2_N/[1|2].
    * COORD1 and COORD2 are the coordinates of the fragment ends.
    * Illumina pair-end reads have read 1 forward and read 2 reverse:
                                   >>>>>>>>>>>>>>
                                                      <<<<<<<<<<<<<<
    * When run in normal (non-wgsim) mode, for pairs where read 1 is from
      the reference strand the coordinates are ordered such that:
      COORD1 < COORD2. For for "flipped" reads where read 1 is from the
      reverse strand the coordinates are ordered such that:
      COORD1 > COORD2.
    * When run in legacy (wgsim) mode, coordinates are always ordered:
      COORD1 < COORD2 and there's no way to tell by inspection what strand
      a read is from."""

    allowed_bases = {'A', 'C', 'G', 'T', 'N'}
    complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N'}
    id_regex = re.compile(
        r"^@(.+)_(\d+)_(\d+)_e(\d+)_e(\d+)_([a-f0-9]+)\/([12])$")

    def __init__(self, filename):

        self.records = []
        self.read_length = -1
        self.forwardized = False
        self.minpos = -1
        self.maxpos = -1

        with open(filename, 'rt') as fh:
            read = frag_start = frag_end = lastline = 0
            for linenum, line in enumerate(fh.readlines(), 1):
                lastline = linenum
                if linenum % 4 == 1:
                    read_id = line.strip()
                    matches = self.id_regex.match(read_id)
                    assert matches, BAD_ID % (linenum, read_id, self.id_regex)
                    frag_start, frag_end = [
                        int(c) for c in matches.groups()[1:3]]
                    read = int(matches.groups()[-1])

                elif linenum % 4 == 2:
                    seq = line.strip()
                    if self.read_length == -1:
                        self.read_length = len(seq)
                    else:
                        assert len(seq) == self.read_length, \
                            BAD_SEQLEN % (linenum, seq, self.read_length)
                    disallowed = set(seq) - self.allowed_bases
                    assert not disallowed, \
                        BAD_BASES % (linenum, disallowed, self.allowed_bases)

                elif linenum % 4 == 3:
                    plus = line.strip()
                    assert plus == "+", BAD_PLUS % (linenum, plus)

                if linenum % 4 == 0:
                    quals = line.strip()
                    assert len(quals) == self.read_length, \
                        BAD_QUALS % (linenum, quals, self.read_length)

                    self.records.append({
                        "id": read_id, "seq": seq, "quals": quals,
                        "frag_start": frag_start, "frag_end": frag_end,
                        "read": read})
                    low = min(frag_start, frag_end)
                    high = max(frag_start, frag_end)
                    if self.minpos == -1 or low < self.minpos:
                        self.minpos = low
                    if self.maxpos == -1 or high > self.maxpos:
                        self.maxpos = high

            assert lastline % 4 == 0, MSG_INCOMPLETE % (filename)

    def coverage(self, pos):
        """Return reads covering pos"""
        # simple logic if all reads are forward on the reference strand:
        if self.forwardized:
            return [r for r in self.records if
                    r['read_start'] <= pos <=
                    r['read_start'] + self.read_length - 1]
        # more cases to consider if not:
        else:
            covering = []
            for r in self.records:
                start = min(r['frag_start'], r['frag_end'])
                end = max(r['frag_start'], r['frag_end'])
                read = r['read']
                flipped = True if r['frag_start'] > r['frag_end'] else False
                if (read == 1 and not flipped and
                        start <= pos <= start + self.read_length - 1 or
                    read == 2 and not flipped and
                        end - self.read_length + 1 <= pos <= end or
                    read == 1 and flipped and
                        end - self.read_length + 1 <= pos <= end or
                    read == 2 and flipped and
                        start <= pos <= start + self.read_length - 1):
                    covering.append(r)
            return covering

    def basecounts(self):
        """Return a dict of { base: count } aggregated over all reads"""
        counts = {}
        for r in self.records:
            for base in r['seq']:
                counts[base] = counts.setdefault(base, 0) + 1
        return counts

    @classmethod
    def forwardize(cls, original):
        """Return a copy of original with all reads turned into forward reads:
           a calculational convenience"""
        fwdized = copy.deepcopy(original)
        for r in fwdized.records:
            frag_start, frag_end = r['frag_start'], r['frag_end']
            read = r['read']
            if (read == 1 and frag_start < frag_end):
                r['read_start'] = frag_start
            elif (read == 1 and frag_start > frag_end):
                r['seq'] = ''.join(cls.revcomp(r['seq']))
                r['quals'] = ''.join(reversed(r['quals']))
                r['read_start'] = frag_start - fwdized.read_length + 1
            elif (read == 2 and frag_start < frag_end):
                r['seq'] = ''.join(cls.revcomp(r['seq']))
                r['quals'] = ''.join(reversed(r['quals']))
                r['read_start'] = frag_end - fwdized.read_length + 1
            elif (read == 2 and frag_start > frag_end):
                r['read_start'] = frag_end
            else:
                raise Exception("Unhandled case:", r)
        fwdized.forwardized = True
        return fwdized

    @classmethod
    def revcomp(cls, seq):
        return [cls.complement[b] for b in reversed(seq)]



'''
Simulate mutations, fragment generation and shotgun sequencing on genomes.
'''
from libc.stdint cimport uint8_t, uint32_t, uint64_t
from libc.math cimport floor, ceil
from libc.stdio cimport FILE, fdopen, stdout, stderr, fprintf
from libc.stdlib cimport calloc, free
from cpython cimport bool
cimport numpy as np


import argparse
import numpy as np
import sys
from datetime import datetime
from xml.etree import ElementTree as ET


cdef extern from "<stdlib.h>":
    void exit(int)
    double drand48()
    void srand48(long)


cdef extern int genreads(FILE *fpout1,
                         FILE *fpout2,
                         uint8_t *s1,
                         uint8_t *s2,
                         uint32_t *rel1,
                         uint32_t *rel2,
                         uint32_t len1,
                         uint32_t len2,
                         uint64_t n_pairs,
                         int dist,
                         int std_dev,
                         int size_l,
                         int size_r,
                         double error_rate,
                         double ambig_frac,
                         const char *seqname,
                         int num_quals,
                         double ***p,
                         char ***q,
                         double **conversions,
                         int wgsim_mode)


cdef uint8_t *nst_nt4_table = [
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 5, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 0, 4, 1,  4, 4, 4, 2,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  3, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 0, 4, 1,  4, 4, 4, 2,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  3, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,
    4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4,  4, 4, 4, 4
]


cdef bool WGSIM_MODE = False


HELP_SOMATIC_MODE = (
    'If "-S, --somatic-mode" is specified then mutation and read generation will '
    'be run /twice/ - the first time generating "germline" mutations, the second '
    'time generating "somatic" mutations. Specifying the other options in this '
    'group has no effect if not in somatic mode.')
HELP_WGSIM_MODE = (
    'In this mode insertions are generated using the same logic as original wgsim.c'
    ' - i.e. max_insertion is set to 4, and insert bases are reversed with respect '
    'to generation order.')
HELP_FRAGMENTS = (
    'The transition/transversion rates represent the chance that the given random '
    'base conversion occurs at any position. This is applied after fragment '
    'generation but before sequencing read error, and can be used to model sample '
    'degradation, e.g. with a non-zero C>T rate for FFPE samples.')
HELP_READS = (
    'If -e is specified then a fixed error rate (and quality string) is used '
    'along the entire read. If -Q is specified then quality scores will be '
    'randomly generated according to the distributions specified in the two '
    'files, and the error rate and quality value is calculated per base. The '
    'files specified by -Q should be qprofiler-like XML documents with <QUAL> '
    'elements containing <Cycle> and <TallyItem> elements with "count" and '
    '"value" attributes.')
MSG_CTOR_SEQ_OR_SIZE = (
    'DipSeq constructor must get a sequence or size, not both.\n')
MSG_NO_SYNC_REF = (
    'A and B sequences of reference do not start at same relative position\n')
MSG_REF_NOT_HAPPY = (
    'A and B sequences of reference are not haploid at %i\'th value\n')
MSG_UNKGT = 'Unhandled genotype at %s %i %s %s %s\n'
MSG_UNKVAR = 'Unhandled variant type %s %i %s %s\n'
MSG_SKIP_MUT = (
    'Mutation at %(POS)i in allele %(allele)i is in deleted region of original, '
    'skipping...\n')
EXCEPT_MUT = (
    'Mutation specified at POS %(POS)i is unsupported, being within a previously '
    'specified deletion at POS %(OLDPOS)i.')
EXCEPT_CYCLENUM = (
    'Cycles found in file: %i, specified read length: %i: cycles must be 1 + '
    'read length')


cdef class VCF:

    cdef public str sample
    cdef public str header
    cdef public list columns
    cdef public list records

    def __cinit__(self, sample='SAMPLE'):
        self.sample = sample
        self.header = (
            "##fileformat=VCFv4.1\n"
            "##INFO=<ID=AF,Number=A,Type=Float,Description=\"Allele Frequency\">\n"
            "##INFO=<ID=SOMATIC,Number=0,Type=Flag,Description=\"Somatic event\">\n"
            "##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n"
            "##FORMAT=<ID=GQ,Number=1,Type=Integer,Description=\"Genotype Quality\">\n"
            "#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t%s\n" %
            self.sample)
        self.columns = ['#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL',
                        'FILTER', 'INFO', 'FORMAT', 'SAMPLE']
        self.records = []

    def tuples(self):
        '''
        Generator returning CHROM, POS, REF, ALT, GT tuples.
        Records are returned ordered by CHROM, POS, REF, ALT so at the same
        position snps are before insertions which are before deletions.
        When all records are exhausted, dummy records are returned for ever
        (CHROM='', POS=0)
        '''
        cdef dict r
        cdef int POS
        cdef str CHROM, REF, ALT, GT
        cdef object iterrec
        self.records.sort(
            key=lambda x: (
                x['#CHROM'],
                x['POS'],
                len(x['REF']),
                len(x['ALT'])))
        iterrec = iter(self.records)
        while True:
            try:
                r = iterrec.__next__()
            except StopIteration:
                r = {}
            CHROM = r.get('#CHROM', '')
            POS = r.get('POS', 0)
            REF = r.get('REF', '')
            ALT = r.get('ALT', '')
            GT = r.get('SAMPLE', '0|0')[:3]
            yield CHROM, POS, REF, ALT, GT

    def write(self, dest):
        '''
        @param dest: file or stream or anything with write and flush.
        '''
        cdef dict r
        dest.write(self.header)
        for r in self.records:
            dest.write('\t'.join(
                map(str, [r.get(col, '') for col in self.columns])) + '\n')

    @classmethod
    def fromfile(cls, filename, sample='SAMPLE'):
        '''
        Return VCF read from a vcf file
        '''
        cdef VCF v
        cdef str line
        cdef list tkns
        cdef dict r
        v = cls(sample)
        with open(filename, 'rt') as fh:
            for line in fh:
                if line.startswith('#'):
                    continue
                tkns = line.strip().split('\t')
                if len(tkns):
                    r = {v.columns[i]: tkns[i] for i in range(len(v.columns))}
                    r['POS'] = int(r['POS'])
                    v.records.append(r)
        return v


cdef class DipSeq:

    cdef public str seqid, description
    cdef public uint8_t fold              # line length for output
    cdef public uint8_t[:] seqA, seqB     # base sequence
    cdef public uint32_t[:] relA, relB    # position on reference
    cdef public uint32_t stopA, stopB

    def __cinit__(self, seqid, description, hapseq=None, size=None, fold=80):
        '''
        Parameterized constructor initialized either with a haploid (1-d) buffer
        as seq argument or empty, if supplied size.
        '''
        cdef uint32_t i
        self.seqid = seqid
        self.description = description
        self.fold = fold

        if hapseq is not None and size is not None:
            raise Exception(MSG_CTOR_SEQ_OR_SIZE)
        elif hapseq is not None:
            self.seqA = self.seqB = hapseq
            self.relA = self.relB = np.empty((self.seqA.shape[0],), 'uint32')
            self.stopA = self.stopB = self.seqA.shape[0]
            self.fold = self.stopA if self.stopA < self.fold else self.fold
            for i in range(self.stopA):
                # bases (A,C,G,T,N,-) into ints (0,1,2,3,4,5)
                self.seqA[i] = self.seqB[i] = nst_nt4_table[self.seqA[i]]
                # relative position is simply 1-based index
                self.relA[i] = self.relB[i] = i + 1
        else:
            self.seqA = np.empty((size,), 'uint8')
            self.seqB = np.empty((size,), 'uint8')
            self.relA = np.empty((size,), 'uint32')
            self.relB = np.empty((size,), 'uint32')
            self.stopA = size
            self.stopB = size

    cdef get_ptrs(self, uint8_t *seq[2], uint32_t *rel[2]):
        seq[0] = &(self.seqA[0])
        seq[1] = &(self.seqB[0])
        rel[0] = &(self.relA[0])
        rel[1] = &(self.relB[0])

    cdef get_stop(self, uint32_t stop[2]):
        stop[0] = self.stopA
        stop[1] = self.stopB

    @classmethod
    def mutagen(cls,
                DipSeq reference not None,
                VCF vcf not None,
                double mut_rate,
                double homo_frac,
                double indel_frac,
                double indel_extend,
                int max_insertion,
                bool somatic_mode=False):
        '''
        Generate mutations on haploid reference sequence, updating vcf in-place.
        '''
        cdef int allele, deleting = 0, appended = 0
        cdef uint8_t refbase, prevref, snp, j
        cdef uint8_t *refseq[2]
        cdef uint32_t i, POS
        cdef uint32_t rstop[2]
        cdef uint32_t *refrel[2]
        cdef str seqid = reference.seqid
        cdef list records = vcf.records, alt
        cdef dict vrec = {}
        cdef dict homrec = {'ID': '.', 'QUAL': '.', 'FILTER': 'PASS', 'INFO': '',
                            'FORMAT': 'GT:GQ', 'SAMPLE': '1|1:200'}
        cdef dict hetrec = {'ID': '.', 'QUAL': '.', 'FILTER': 'PASS', 'INFO': '',
                            'FORMAT': 'GT:GQ'}

        if somatic_mode:
            homrec['INFO'] += 'SOMATIC;'
            hetrec['INFO'] += 'SOMATIC;'

        reference.get_ptrs(refseq, refrel)
        reference.get_stop(rstop)

        if refrel[0][0] != refrel[1][0]:
            raise Exception(MSG_NO_SYNC_REF)

        # start mutating at second base for convenience so we don't have to handle
        # vcf requirement that REF be _after_ the variant when variant is at POS=1
        prevref = refseq[0][0]

        for i in range(1, rstop[0]):

            if (refseq[0][i] != refseq[1][i] or
                refrel[0][i] != refrel[1][i] or
                refrel[0][i] != i + 1):
                raise Exception(MSG_REF_NOT_HAPPY % i)

            POS = i + 1
            refbase = refseq[0][i]

            if deleting:
                if drand48() < indel_extend:
                    vrec['REF'] += "ACGTN"[refbase]
                    prevref = refbase
                    continue
                else:
                    records.append(vrec)
                    appended = 1
                    deleting = 0

            if refbase < 4 and drand48() < mut_rate:
                vrec = {
                    '#CHROM': seqid,
                    'POS': POS,
                    'REF': "ACGT"[refbase],
                    'ALT': ''}
                appended = 0

                if drand48() >= indel_frac:                         # =SNP=
                    snp = (refbase + <uint8_t>(drand48() * 3.0 + 1)) & 3
                    vrec['ALT'] = "ACGT"[snp]
                    if drand48() < homo_frac:                       # hom SNP
                        vrec.update(homrec)
                    else:                                           # het SNP
                        allele = 0 if drand48() < 0.5 else 1
                        hetrec['SAMPLE'] = '%s|%s:200' % (1 - allele, allele)
                        vrec.update(hetrec)
                    records.append(vrec)
                    appended = 1

                else:
                    if drand48() < 0.5:                             # =DEL=
                        deleting = 1
                        vrec.update({'POS': POS - 1,
                                     'REF': "ACGTN"[prevref] + "ACGT"[refbase],
                                     'ALT': "ACGTN"[prevref]})
                        if drand48() < homo_frac:                   # hom DEL
                            vrec.update(homrec)
                        else:                                       # het DEL
                            allele = 0 if drand48() < 0.5 else 1
                            hetrec['SAMPLE'] = '%s|%s:200' % (
                                1 - allele, allele)
                            vrec.update(hetrec)

                    else:                                           # =INS=
                        alt = insertion(refbase, indel_extend, max_insertion)
                        vrec['ALT'] = ''.join(["ACGT"[j] for j in alt])
                        if drand48() < homo_frac:                   # hom INS
                            vrec.update(homrec)
                        else:                                       # het INS
                            allele = 0 if drand48() < 0.5 else 1
                            hetrec['SAMPLE'] = '%s|%s:200' % (
                                1 - allele, allele)
                            vrec.update(hetrec)
                        records.append(vrec)
                        appended = 1

            prevref = refbase

        # any unfinished deletion
        if vrec.get('#CHROM') and not appended:
            records.append(vrec)

    def transform(self,
                  DipSeq original not None,
                  VCF vcf):
        '''
        Apply mutations from a vcf to a diploid sequence, updating self in place.
        For insertions and deletions, the REF and ALT sequences respectively
        must be a single base only.
        '''
        # gt is coded phased genotype 0=(0|0), 1=(1|0), 2=(0|1), 3=(1|1)
        cdef int allele, c, snp, ins, gt = 0, refsz, altsz
        cdef uint8_t *origseq[2]
        cdef uint8_t *mutseq[2]
        cdef uint32_t POS, OLDPOS, del_l, del_r, r
        cdef uint32_t opos[2]
        cdef uint32_t mpos[2]
        cdef uint32_t ostop[2]
        cdef uint32_t mstop[2]
        cdef uint32_t *origrel[2]
        cdef uint32_t *mutrel[2]
        cdef str alt, seqid = original.seqid.split('.')[0]
        cdef str CHROM, REF, ALT, GT
        cdef object mutations

        original.get_ptrs(origseq, origrel)
        original.get_stop(ostop)
        self.get_ptrs(mutseq, mutrel)
        self.get_stop(mstop)

        # apply mutations to two original alleles independently
        opos[0] = opos[1] = 0
        mpos[0] = mpos[1] = 0
        for allele in range(2):

            # at given location tuples() returns snps, insertions then
            # deletions
            mutations = vcf.tuples()
            while True:
                # Obtain first mutation to apply by advancing until we hit the
                # first VCF record where CHROM matches seqid, or all VCF
                # records are exhausted. The body of the transform method that
                # follows assumes VCF records are ordered by POS for a given
                # CHROM, but no particular order for CHROM blocks.
                CHROM, POS, REF, ALT, GT = mutations.__next__()
                if CHROM in (seqid, ''):
                    break

            del_l = del_r = 0
            OLDPOS = 0
            while opos[allele] < ostop[allele]:

                c = origseq[allele][opos[allele]]
                r = origrel[allele][opos[allele]]

                # in deletion
                if (del_l <= <uint32_t>r <= del_r):
                    if r == POS:
                        # DipSeq.mutagen() doesn't generate mutations within
                        # deletions with the exception that the reference base
                        # of an indel may be the last base of previous deletion
                        if not (len(REF) != len(ALT) and POS == del_r):
                            raise Exception(EXCEPT_MUT %
                                            {'POS': POS, 'OLDPOS': OLDPOS})
                    else:
                        opos[allele] += 1
                        continue

                # process any mutation at this location
                if CHROM == seqid and POS == r:
                    if GT == '1|0':                      # het (1|0)
                        gt = 1
                    elif GT == '0|1':                    # het (0|1)
                        gt = 2
                    elif GT == '1|1':                    # hom
                        gt = 3
                    else:
                        raise Exception(
                            MSG_UNKGT %
                            (seqid, opos[allele], REF, ALT, GT))

                    refsz, altsz = len(REF), len(ALT)
                    if refsz == altsz == 1:              # snp
                        snp = nst_nt4_table[ord(ALT[0])]
                        if gt == 3 or gt == allele + 1:  # hit allele
                            update(allele, mutseq, snp, mutrel, r, mpos, mstop)
                        else:
                            update(allele, mutseq, c, mutrel, r, mpos, mstop)

                    elif altsz > refsz and refsz == 1:   # insertion
                        if ((del_l <= r <= del_r) or POS == OLDPOS):
                            # continued deletion or collocated snp
                            ALT = ALT[1:]
                        if gt == 3 or gt == allele + 1:  # hit allele
                            for alt in ALT:
                                ins = nst_nt4_table[ord(alt)]
                                update(
                                    allele, mutseq, ins, mutrel, r, mpos, mstop)
                        elif not ((del_l <= r <= del_r) or POS == OLDPOS):
                            update(allele, mutseq, c, mutrel, r, mpos, mstop)

                    elif altsz < refsz and altsz == 1:   # deletion
                        if not ((del_l <= r <= del_r) or POS == OLDPOS):
                            # no continued deletion or collocated snp/insertion
                            update(allele, mutseq, c, mutrel, r, mpos, mstop)
                        if gt == 3 or gt == allele + 1:  # hit allele
                            del_l = POS + 1
                            del_r = POS + len(REF) - 1

                    OLDPOS = POS
                    CHROM, POS, REF, ALT, GT = mutations.__next__()
                    # keep processing at same location
                    if POS == OLDPOS:
                        continue
                    else:
                        opos[allele] += 1

                else:
                    update(allele, mutseq, c, mutrel, r, mpos, mstop)
                    opos[allele] += 1

                # get another mutation if new location is past current POS
                while CHROM == seqid and origrel[allele][
                        opos[allele]] > POS and POS > 0:
                    sys.stderr.write(MSG_SKIP_MUT %
                                     {'allele': allele, 'POS': POS})
                    CHROM, POS, REF, ALT, GT = mutations.__next__()

        self.stopA = mpos[0]
        self.stopB = mpos[1]

    def write(self, dest):
        '''
        Output sequences from a DipSeq instance.
        Only the last digit of the reference-relative coordinate is shown.

        @param dest: file or stream or anything with write and flush.
        '''
        cdef int startln, allele, j, fold = self.fold
        cdef uint8_t *seq[2]
        cdef uint32_t i
        cdef uint32_t stop[2]
        cdef uint32_t *rel[2]

        self.get_ptrs(seq, rel)
        self.get_stop(stop)

        for allele in range(2):
            startln = 0
            dest.write('>%s.%s %s\n' % (self.seqid, allele, self.description))
            for i in range(stop[allele]):
                dest.write("ACGTN"[seq[allele][i]])
                if i and not (rel[allele][i]) % fold:
                    dest.write('\n')
                    for j in range(startln, i + 1):
                        dest.write(str(rel[allele][j] % 10))
                    dest.write('\n')
                    startln = i + 1
        dest.flush()


cdef inline int update(int allele,
                       uint8_t *seq[2],
                       uint8_t seqval,
                       uint32_t *rel[2],
                       uint32_t relval,
                       uint32_t *pos,
                       uint32_t *stop) nogil:
    '''
    Helper to update DipSeq sequences and position marker.
    '''
    if pos[allele] >= stop[allele]:
        fprintf(
            stderr,
            "Array bounds exceeded. Insertions too large/too many? "
            "Try smaller values of -X, -r, and/or -R")
        return 1
    seq[allele][pos[allele]] = seqval
    rel[allele][pos[allele]] = relval
    pos[allele] += 1
    return 0


cdef insertion(uint8_t first,
               double indel_extend,
               int max_insertion):
    '''
    Build and return list of randomly inserted characters as bytes in vcf 'ALT'
    format i.e. with first character as reference base. If WGSIM_MODE is set,
    the generated insertion will be reversed with respect to generation order.
    '''
    cdef int num_ins = 0
    cdef list ins
    ins = [] if WGSIM_MODE else [first, ]
    while True:
        num_ins += 1
        ins.append(<uint8_t>(drand48() * 4.0))
        if num_ins >= max_insertion or drand48() > indel_extend:
            if WGSIM_MODE:
                # reverse for consistency with bitshift method
                ins.append(first)
                return ins[::-1]
            else:
                return ins


def read_fasta(filename):
    '''
    Read from file, yield FASTA sequences.
    '''
    cdef str seqid = '', description = ''
    cdef object seqb = bytearray()
    cdef int fold = 80
    cdef bool firstseq = True
    cdef bool firstseqline = True

    with open(filename, 'rb') as fh:

        for line in fh:
            l = line[:-1]           # strip newline
            if l == b'':            # blank line
                continue
            elif l[0] in (59, 62):  # ';', '>'
                f = DipSeq(seqid, description, seqb, fold=fold)
                tkns = l.decode('ASCII').lstrip('>').split(' ', 1)
                seqid = tkns[0] if len(tkns) >= 1 else ''
                description = tkns[1] if len(tkns) == 2 else ''
                seqb = bytearray()
                firstseqline = True
                if firstseq:
                    firstseq = False
                else:
                    yield f
            else:
                if firstseqline:
                    firstseqline = False
                    fold = len(l)
                seqb.extend(l)
        # EOF
        yield(DipSeq(seqid, description, seqb, fold=fold))


def gen_quals(str filename, int readlen, int num_quals, char[:, :] qvals, double[:, :] pvals):
    '''
    Generate quality values from a file
    '''
    cdef int i, j
    cdef object doc, root, Q, cumdist
    cdef list cycles, tallies
    cdef np.ndarray qarray

    doc = ET.parse(filename)
    root = doc.getroot()
    Q = root.find('.//QUAL')
    cycles = Q.findall('.//Cycle')

    if len(cycles) != readlen + 1:
        raise Exception(EXCEPT_CYCLENUM % (len(cycles), readlen))

    tallies = []
    for c in cycles:
        cyclenum = int(c.get('value'))
        ctallies = sorted(
            (cyclenum, int(t.get('value')), int(t.get('count')))
            for t in c.findall('TallyItem'))
        tallies.extend(ctallies)

    cumdist = {}
    for cycle, qual, count in tallies:
        cumdist.setdefault(cycle, [])
        prev = cumdist[cycle][-1][1] if len(cumdist[cycle]) else 0
        new = prev + count
        cumdist[cycle].append((qual, new))

    for i in range(num_quals):
        for j in range(readlen):
            qvals[i, j] = _randqual(cumdist, j + 1)
            pvals[i, j] = 10 ** (qvals[i, j] / -10.0)  # Phred definition


cdef inline int _randqual(dict dist, int cycle):
    '''
    Return a random, representative quality value for a cycle given a dict of
    lists of tuples representing cumulative distribution function of qualities
    per cycle:{ <cycle1>: [(<qual1>, <cumulative count1>),
                           (<qual2>, <cumulative count2>),
                           ...],
    '''
    cdef long maxcount, count, rand
    cdef int qual
    maxcount = dist[cycle][-1][1]
    rand = <long>(drand48() * maxcount)
    for qual, count in dist[cycle]:
        if count > rand:
            return qual
    return 0


def _t_randqual(dist, cycle):
    '''
    Shim for unit testing the cdef inline method _randqual
    '''
    return _randqual(dist, cycle)


def reseed(seed):
    '''
    Hook for (re)seeding the random number generator
    '''
    srand48(seed)


def get_args(argv):
    '''
    Return options and arguments
    '''
    class Range(object):

        def __init__(self, start, end):
            self.start = start
            self.end = end

        def __eq__(self, other):
            return self.start <= other <= self.end

        def __repr__(self):
            return "Range %s<=<%s" % (str(self.start), str(self.end))

    p = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument('fasta', help='Reference FASTA')
    p.add_argument('read1fq', help='Output file for read1', type=argparse.FileType('wb'))
    p.add_argument('read2fq', help='Output file for read2', type=argparse.FileType('wb'))

    mutgrp = p.add_argument_group('Mutations')
    mutgrp.add_argument('-r', '--mut-rate', help='mutation rate', type=float, default=0.001, choices=[Range(0.0,1.0)])
    mutgrp.add_argument('-H', '--homo-frac', help='fraction of mutations that are homozygous', type=float, default=0.333333, choices=[Range(0.0,1.0)])
    mutgrp.add_argument('-R', '--indel-frac', help='fraction of mutations that are indels', type=float, default=0.15, choices=[Range(0.0,1.0)])
    mutgrp.add_argument('-X', '--indel-extend', help='probability an indel is extended', type=float, default=0.3, choices=[Range(0.0,1.0)])
    mutgrp.add_argument('-M', '--max-insertion', help='Maximum size of generated insertions (regardless of -X value)', type=int, default=1000)
    mutgrp.add_argument('-n', '--sample-name', help='name of sample for vcf output', type=str, default='SAMPLE')

    mutgrpio = mutgrp.add_mutually_exclusive_group(required=True)
    mutgrpio.add_argument('-o', '--output', metavar='VCF', help='output generated mutations to file', type=argparse.FileType('wt'))
    mutgrpio.add_argument('-V', '--vcf-input', help='use input vcf file as source of mutations instead of randomly generating them. VCF records must be grouped by CHROM and ordered by POS within each CHROM.', type=str)

    mutgrp2 = p.add_argument_group(title='Somatic mutations', description=HELP_SOMATIC_MODE)
    mutgrp2.add_argument('-S', '--somatic-mode', action='store_true')
    mutgrp2.add_argument('--mut-rate2', help='somatic mutation rate', type=float, default=0.000001, choices=[Range(0.0,1.0)])
    mutgrp2.add_argument('--homo-frac2', help='fraction of somatic mutations that are homozygous', type=float, default=0.333333, choices=[Range(0.0,1.0)])
    mutgrp2.add_argument('--indel-frac2', help='fraction of somatic mutations that are indels', type=float, default=0.15, choices=[Range(0.0,1.0)])
    mutgrp2.add_argument('--indel-extend2', help='probability a somatic indel is extended', type=float, default=0.3, choices=[Range(0.0,1.0)])
    mutgrp2.add_argument('--max-insertion2', help='Maximum size of generated somatic insertions (regardless of -X value)', type=int, default=1000)
    mutgrp2.add_argument('--contamination', help='fraction of reads generated from "germline" sequence', type=float, default=0.0, choices=[Range(0.0,1.0)])
    mutgrp2.add_argument('--sample-name2', help='name of sample for vcf2 output', type=str, default='SOMATIC')

    mutgrp2io = mutgrp2.add_mutually_exclusive_group()
    mutgrp2io.add_argument('--output2', metavar='VCF2', help='output generated somatic mutations to file', type=argparse.FileType('wt'))
    mutgrp2io.add_argument('--vcf-input2', help='use input vcf file as source of somatic mutations instead of randomly generating them. VCF records must be grouped by CHROM and ordered by POS within each CHROM.', type=str)

    fragrp = p.add_argument_group('Fragments', description=HELP_FRAGMENTS)
    fragrp.add_argument('-z', '--size', help='mean fragment size', type=int, default=500)
    fragrp.add_argument('-s', '--std-dev', help='fragment standard deviation', type=int, default=50)
    fragrp.add_argument('--AC', help='A>C transversion rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--AG', help='A>G transition rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--AT', help='A>T transversion rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--CA', help='C>A transversion rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--CG', help='C>G transversion rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--CT', help='C>T transition rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--GA', help='G>A transition rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--GC', help='G>C transversion rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--GT', help='G>T transversion rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--TA', help='T>A transversion rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--TC', help='T>C transition rate', type=float, choices=[Range(0.0,1.0)])
    fragrp.add_argument('--TG', help='T>G transversion rate', type=float, choices=[Range(0.0,1.0)])

    rdsgrp = p.add_argument_group('Reads', description=HELP_READS)
    rdsgrp.add_argument('-N', '--num-pairs', help='number of read pairs', type=int, default=1000000)
    rdsgrp.add_argument('-1', '--length1', help='length of read 1', type=int, default=100)
    rdsgrp.add_argument('-2', '--length2', help='length of read 2', type=int, default=100)
    rdsgrp.add_argument('-A', '--ambig-frac', help='discard read if fraction of "N" bases exceeds this', type=float, default=1.0, choices=[Range(0.0,1.0)])

    rdsgrperr = rdsgrp.add_mutually_exclusive_group()
    rdsgrperr.add_argument('-e', '--error-rate', help='read error rate (constant)', type=float, default=0.002, choices=[Range(0.0,1.0)])
    rdsgrperr.add_argument('-Q', '--quals-from', help='generate random quality strings for read 1 and read 2 respectively from the distributions specified in the files', type=str, nargs=2, metavar=('R1_QUALS', 'R2_QUALS'))
    # --num-quals is part of rdsgrp but leave it here after --quals-from
    # so it's grouped logically in -h, --help output
    rdsgrp.add_argument('--num-quals', help='number of quality strings to generate from distribution files', type=int, default=10000)

    othgrp = p.add_argument_group('Other')
    othgrp.add_argument('-d', '--seed', help='seed for random generator (default=current time)', type=int, default=datetime.now().strftime('%s'))
    othgrp.add_argument('-t', '--test-output', help='print mutated sequences to stdout', action='store_true')
    othgrp.add_argument('-w', '--wgsim-mode', help=HELP_WGSIM_MODE, action='store_true')

    args = p.parse_args(argv)

    global WGSIM_MODE
    WGSIM_MODE = args.wgsim_mode
    if WGSIM_MODE:
        args.max_insertion = 4
        args.max_insertion2 = 4
    del args.wgsim_mode

    if args.somatic_mode and not args.vcf_input2 and not args.output2:
        p.error(
            'one of --output2 or --vcf-input2 is required when specifying somatic mode')

    return args


def workflow(args):
    '''
    Run the workflow specified by the command line args.
    '''

    # variables from command line args
    cdef str fasta = args.fasta, sample_name = args.sample_name, \
        sample_name2 = args.sample_name2, vcf_input = args.vcf_input, \
        vcf_input2 = args.vcf_input2, r1_quals, r2_quals
    cdef object read1fq = args.read1fq, read2fq = args.read2fq, \
        output = args.output, output2 = args.output2,
    cdef int max_insertion = args.max_insertion, size = args.size, \
        std_dev = args.std_dev, num_pairs = args.num_pairs, \
        length1 = args.length1, length2 = args.length2, \
        num_quals = args.num_quals, seed = args.seed, \
        max_insertion2 = args.max_insertion2
    cdef double mut_rate = args.mut_rate, homo_frac = args.homo_frac, \
        indel_frac = args.indel_frac, indel_extend = args.indel_extend, \
        mut_rate2 = args.mut_rate2, homo_frac2 = args.homo_frac2, \
        indel_frac2 = args.indel_frac2, indel_extend2 = args.indel_extend2, \
        contamination = args.contamination, error_rate = args.error_rate, \
        ambig_frac = args.ambig_frac
    cdef bool somatic_mode = args.somatic_mode, test_output = args.test_output

    cdef uint64_t tot_len = 0, n_pairs, n_som, n_grm
    cdef int n_ref = 0, i, mutseqsize, mutseq2size
    cdef VCF vcf, vcf2
    cdef DipSeq refseq, mutseq, mutseq2
    cdef FILE *fpout1
    cdef FILE *fpout2
    cdef uint8_t[:, :, ::1] qvals
    cdef double[:, :, ::1] pvals
    cdef double *conversions[4]
    cdef char **q[2]
    cdef double **p[2]

    sys.stderr.write("[%s] seed = %i\n" % (__name__, seed))
    reseed(seed)

    fpout1 = fdopen(read1fq.fileno(), 'wb')
    fpout2 = fdopen(read2fq.fileno(), 'wb')

    if vcf_input:
        sys.stderr.write(
            "[%s] reading mutations from file %s (1)\n" %
            (__name__, vcf_input))
        vcf = VCF.fromfile(vcf_input, sample_name)
    else:
        vcf = VCF(sample_name)

    if somatic_mode:
        if vcf_input2:
            sys.stderr.write(
                "[%s] reading mutations from file %s (2)\n" %
                (__name__, vcf_input2))
            vcf2 = VCF.fromfile(vcf_input2, sample_name2)
        else:
            vcf2 = VCF(sample_name2)

    if args.quals_from:
        r1_quals, r2_quals = args.quals_from
        qvals = np.ndarray((2, num_quals, max(length1, length2)), order='C', dtype='u1')
        pvals = np.ndarray((2, num_quals, max(length1, length2)), order='C')
        for i, (quals_file, read_length) in \
            enumerate(zip([r1_quals, r2_quals], [length1, length2])):
            sys.stderr.write(
                "[%s] generating qualities for read %s from %s\n" %
                (__name__, i + 1, quals_file))
            reseed(seed)
            gen_quals(quals_file, read_length, num_quals, qvals[i], pvals[i])
            q[i] = <char**>calloc(num_quals, sizeof(char*))
            p[i] = <double**>calloc(num_quals, sizeof(double*))
            for j in range(num_quals):
                q[i][j] = <char*>&qvals[i, j, 0]
                p[i][j] = <double*>&pvals[i, j, 0]
    else:
        # special value to indicate to genreads() to use fixed error rate
        num_quals = 0

    for i in range(4):                                                          
        conversions[i] = <double*>calloc(4, sizeof(double))   
    # use -1 as unambiguous 'not set' value
    conversions[0][1] = -1 if args.AC is None else args.AC
    conversions[0][2] = -1 if args.AG is None else args.AG
    conversions[0][3] = -1 if args.AT is None else args.AT
    conversions[1][0] = -1 if args.CA is None else args.CA
    conversions[1][2] = -1 if args.CG is None else args.CG
    conversions[1][3] = -1 if args.CT is None else args.CT
    conversions[2][0] = -1 if args.GA is None else args.GA
    conversions[2][1] = -1 if args.GC is None else args.GC
    conversions[2][3] = -1 if args.GT is None else args.GT
    conversions[3][0] = -1 if args.TA is None else args.TA
    conversions[3][1] = -1 if args.TC is None else args.TC
    conversions[3][2] = -1 if args.TG is None else args.TG

    for refseq in read_fasta(fasta):
        n_ref += 1
        tot_len += refseq.stopA
    sys.stderr.write("[%s] %d input sequences, total length: %i\n" %
                     (__name__, n_ref, tot_len))

    # iterate through input sequences
    for refseq in read_fasta(fasta):
        n_pairs = <uint64_t>(<float>refseq.stopA / <float>tot_len * num_pairs)
        n_grm = <uint64_t>floor(n_pairs * contamination) if somatic_mode else n_pairs
        n_som = <uint64_t>ceil(n_pairs * (1 - contamination))

        if refseq.stopA < <uint32_t>(size + 3 * std_dev):
            sys.stderr.write(
                "[%s] skip sequence '%s' as it is shorter than %d\n" %
                (__name__, refseq.seqid, size + 3 * std_dev))
            continue

        # this size is a bit arbitrary, but seems to cope with most cases
        mutseqsize = <int>(1.1 * refseq.seqA.shape[0] + 10 * max_insertion)
        mutseq = DipSeq(refseq.seqid + '.1',
                        refseq.description + '.1',
                        size=mutseqsize,
                        fold=refseq.fold)

        if not vcf_input:
            sys.stderr.write("[%s] generating mutations (1)\n" % (__name__,))
            DipSeq.mutagen(refseq,
                           vcf,
                           mut_rate,
                           homo_frac,
                           indel_frac,
                           indel_extend,
                           max_insertion)
        mutseq.transform(refseq, vcf)
        if test_output:
            mutseq.write(sys.stdout)

        # sample "germline" sequence
        bseqid = bytes(mutseq.seqid, 'ASCII')
        sys.stderr.write("[%s] generating %i reads from sequence %s\n" %
                         (__name__, n_grm, mutseq.seqid))
        if n_grm:
            # reseed(seed) # reseed during testing
            genreads(fpout1, fpout2, &(mutseq.seqA[0]), &(mutseq.seqB[0]),
                     &(mutseq.relA[0]), &(mutseq.relB[0]),
                     mutseq.stopA, mutseq.stopB,
                     n_grm, size, std_dev, length1, length2,
                     error_rate, ambig_frac, bseqid, num_quals, p, q,
                     conversions, int(WGSIM_MODE))

        if somatic_mode:
            mutseq2size = <int>(1.1 * mutseq.seqA.shape[0] + 10 * max_insertion2)
            mutseq2 = DipSeq(refseq.seqid + '.2',
                             refseq.description + '.2',
                             size=mutseq2size,
                             fold=refseq.fold)

            if not vcf_input2:
                sys.stderr.write(
                    "[%s] generating mutations (2)\n" %
                    (__name__,))
                DipSeq.mutagen(refseq,
                               vcf2,
                               mut_rate2,
                               homo_frac2,
                               indel_frac2,
                               indel_extend2,
                               max_insertion2,
                               True)
            mutseq2.transform(mutseq, vcf2)
            if test_output:
                mutseq2.write(sys.stdout)

            # sample somatic sequence
            bseqid = bytes(mutseq2.seqid, 'ASCII')
            sys.stderr.write("[%s] generating %i reads from sequence %s\n" %
                             (__name__, n_som, mutseq2.seqid))
            if n_som:
                # reseed(seed) # reseed during testing
                genreads(fpout1, fpout2, &(mutseq2.seqA[0]), &(mutseq2.seqB[0]),
                         &(mutseq2.relA[0]), &(mutseq2.relB[0]),
                         mutseq2.stopA, mutseq2.stopB,
                         n_som, size, std_dev, length1, length2,
                         error_rate, ambig_frac, bseqid, num_quals, p, q,
                         conversions, int(WGSIM_MODE))

    if not vcf_input:
        vcf.write(output)
    if somatic_mode and not vcf_input2:
        vcf2.write(output2)

    if output:
        output.close()
    if output2:
        output2.close()

    read1fq.close()
    read2fq.close()

    if args.quals_from:
        for i in range(2):
            free(q[i])
            free(p[i])

    return 0



"""Tests for qasim package"""
import sys
import tempfile
import unittest

from contextlib import contextmanager
from io import StringIO
from os.path import join as path_join
from os.path import dirname
from xml.etree import ElementTree as ET

import numpy as np

from qasim import qasim
from qasim.qasim import EXCEPT_MUT, MSG_SKIP_MUT, MSG_CTOR_SEQ_OR_SIZE
from qasim.qasim import VCF, DipSeq
from .testutil import Fastq


@contextmanager
def captured_output():
    """
    https://stackoverflow.com/a/17981937/6705037
    """
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


def base(code):
    """
    Return letter code for integer encoded base
    """
    return "ACGTN-"[code]


class TestVcf(unittest.TestCase):
    """VCF class tests"""

    @classmethod
    def setUpClass(cls):
        """Load some test resources"""
        # Notify aggressively so we know where we are if setup fails
        sys.stdout.write("\nTestVcf.setUpClass\n")
        sys.stdout.flush()
        cls.vcf0 = VCF.fromfile(
            path_join(dirname(__file__), 'resources/test.vcf.0'))

    def test_ctor(self):
        """Test the no-args constructor"""
        vcf = VCF()
        self.assertEqual(vcf.sample, "SAMPLE")
        self.assertEqual(len(vcf.records), 0)

    def test_tuples(self):
        """Test the VCF.tuples() method"""
        CHROM, POS, REF, ALT, GT = next(self.vcf0.tuples())
        self.assertEqual(CHROM, "TEST")
        self.assertEqual(POS, 5)
        self.assertEqual(REF, "C")
        self.assertEqual(ALT, "G")
        self.assertEqual(GT, "0|1")

    def test_write(self):
        """Test the VCF.write() method"""
        with tempfile.TemporaryFile(mode='w+t') as fh:
            self.vcf0.write(fh)
            fh.seek(0)
            lines = fh.readlines()
            self.assertEqual(''.join(lines[:6]), self.vcf0.header)
            self.assertEqual(
                ''.join(lines[6:]), '\n'.join(
                    '\t'.join(str(r[c]) for c in self.vcf0.columns)
                    for r in self.vcf0.records) + '\n')

    def test_fromfile(self):
        """Test the VCF.fromfile() method"""
        vcf = VCF.fromfile(
            path_join(dirname(__file__), 'resources/test.vcf.0'), "sample1")
        self.assertEqual(vcf.sample, "sample1")
        self.assertEqual(len(vcf.records), 3)


class TestDipSeq(unittest.TestCase):
    """DipSeq class tests"""

    @classmethod
    def setUpClass(cls):
        """Load some test resources"""
        # Notify aggressively so we know where we are if setup fails
        sys.stdout.write("\nTestDipSeq.setUpClass\n")
        sys.stdout.flush()
        cls.mut_seq = path_join(
            dirname(__file__), 'resources/mutagen_result.sequence')
        cls.mut_vcf = path_join(
            dirname(__file__), 'resources/mutagen_result.vcf')
        cls.fa0 = next(qasim.read_fasta(
            path_join(dirname(__file__), 'resources/test.fa.0')))
        cls.fa1 = next(qasim.read_fasta(
            path_join(dirname(__file__), 'resources/test.fa.1')))
        cls.vcf0 = VCF.fromfile(
            path_join(dirname(__file__), 'resources/test.vcf.0'))
        cls.vcf1 = VCF.fromfile(
            path_join(dirname(__file__), 'resources/test.vcf.1'))
        cls.vcf2 = VCF.fromfile(
            path_join(dirname(__file__), 'resources/test.vcf.2'))
        cls.vcfgrm = VCF.fromfile(
            path_join(dirname(__file__), 'resources/test.vcf.3_1'))
        cls.vcfsom = VCF.fromfile(
            path_join(dirname(__file__), 'resources/test.vcf.3_2'))
        cls.vcf4 = VCF.fromfile(
            path_join(dirname(__file__), 'resources/test.vcf.4'))

    def test_ctor_from_size(self):
        """Test the constructor that takes size argument"""
        d = DipSeq("T", "TEST", size=6)
        self.assertEqual(d.stopA, 6)
        self.assertEqual(d.stopB, 6)

    def test_ctor_from_seq(self):
        """Test the constructor that takes a sequence argument"""
        d = DipSeq("T", "TEST", hapseq=bytearray([65, 67, 71, 84, 78, 45]))
        self.assertEqual(d.stopA, 6)
        self.assertEqual(d.stopB, 6)
        self.assertEqual(list(d.seqA), [0, 1, 2, 3, 4, 5])

    def test_ctor_seq_and_size(self):
        """Test Exception is correctly raised for bad ctor args"""
        with self.assertRaisesRegex(Exception, MSG_CTOR_SEQ_OR_SIZE):
            DipSeq("T", "TEST", bytearray([65]), 1)

    def test_write(self):
        """Test the DipSeq.write() method"""
        d = DipSeq("T", "TEST", hapseq=bytearray([65, 67, 71, 84, 78]))
        out = StringIO()
        d.write(out)
        self.assertEqual(out.getvalue(), (
            ">T.0 TEST\n"
            "ACGTN\n"
            "12345\n"
            ">T.1 TEST\n"
            "ACGTN\n"
            "12345\n"))

    def test_mutagen(self):
        """tests both mutagen and transform methods"""
        # Another case where we take a shortcut by simply asserting that the
        # data in the test resource files mut_vcf and mut_seq are valid "by
        # inspection" and require that the test run matches them every time.
        # In this case cross-referencing the randomly generated mutations in
        # the vcf with the transformed sequence file is what's required. You
        # should eyeball it yourself — it's interesting!
        refseq = self.fa1
        vcf = VCF("sample1")
        mut_rate = 0.01
        homo_frac = 0.333333
        indel_frac = 0.15
        indel_extend = 0.3
        max_insertion = 1000
        qasim.reseed(12345678)  # deterministic iff we set the seed
        DipSeq.mutagen(refseq, vcf, mut_rate, homo_frac, indel_frac,
                       indel_extend, max_insertion)
        out = StringIO()
        vcf.write(out)
        with open(self.mut_vcf) as fh:
            self.assertEqual(out.getvalue(), ''.join(fh.readlines()))
        mutseq = DipSeq(
            refseq.seqid + '.mut',
            refseq.description,
            size=refseq.seqA.shape[0] * 2,
            fold=refseq.fold)
        mutseq.transform(refseq, vcf)
        out = StringIO()
        mutseq.write(out)
        with open(self.mut_seq) as fh:
            self.assertEqual(out.getvalue(), ''.join(fh.readlines()))

    def test_transform_0(self):
        """germline het & hom snps"""
        refseq = self.fa0
        mutseq = DipSeq(
            refseq.seqid + '.mut',
            refseq.description,
            size=refseq.seqA.shape[0] * 2,
            fold=refseq.fold)
        mutseq.transform(refseq, self.vcf0)
        for i in range(refseq.stopA):
            POS = i + 1               # VCF coords
            if POS == 5:
                self.assertEqual(base(mutseq.seqA[i]), base(refseq.seqA[i]))
                self.assertEqual(base(mutseq.seqB[i]), "G")
            elif POS == 9:
                self.assertEqual(base(mutseq.seqA[i]), "G")
                self.assertEqual(base(mutseq.seqB[i]), base(refseq.seqA[i]))
            elif POS == 13:
                self.assertEqual(base(mutseq.seqA[i]), "T")
                self.assertEqual(base(mutseq.seqA[i]), base(mutseq.seqB[i]))
                self.assertNotEqual(base(mutseq.seqA[i]), base(refseq.seqA[i]))
            else:
                self.assertEqual(base(mutseq.seqA[i]), base(refseq.seqA[i]))
                self.assertEqual(base(mutseq.seqB[i]), base(refseq.seqA[i]))

    def test_transform_1(self):
        """simple het & hom indels"""
        refseq = self.fa0
        mutseq = DipSeq(
            refseq.seqid + '.mut',
            refseq.description,
            size=refseq.seqA.shape[0] * 2,
            fold=refseq.fold)
        mutseq.transform(refseq, self.vcf1)
        insertion_1 = range(4, 7)
        self.assertEqual(
            ''.join(base(mutseq.seqA[POS - 1]) for POS in insertion_1), "AGG")
        self.assertEqual(
            list(mutseq.relA[POS - 1] for POS in insertion_1), [4, 4, 4])
        insertion_2 = range(5, 9)
        self.assertEqual(
            ''.join(base(mutseq.seqB[POS - 1]) for POS in insertion_2), "CTTT")
        self.assertEqual(
            list(mutseq.relB[POS - 1] for POS in insertion_2), [5, 5, 5, 5])
        deletion_1A = range(15, 17)
        deletion_1B = range(16, 18)
        self.assertEqual(
            ''.join(base(mutseq.seqA[POS - 1]) for POS in deletion_1A), "CC")
        self.assertEqual(
            list(mutseq.relA[POS - 1] for POS in deletion_1A), [13, 16])
        self.assertEqual(
            ''.join(base(mutseq.seqB[POS - 1]) for POS in deletion_1B), "CC")
        self.assertEqual(
            list(mutseq.relB[POS - 1] for POS in deletion_1B), [13, 16])

    def test_transform_2(self):
        """complex overlapping mutations"""
        refseq = self.fa0
        mutseq = DipSeq(
            refseq.seqid + '.mut',
            refseq.description,
            size=refseq.seqA.shape[0] * 2,
            fold=refseq.fold)
        mutseq.transform(refseq, self.vcf2)
        out = StringIO()
        mutseq.write(out)
        # we take a bit of a shorcut and rather than testing all the
        # explicit logic of the transformation we just test equality to
        # this output that we assert is valid "by inspection".
        self.assertEqual(out.getvalue(), (
            ">TEST.mut.0 small fasta for testing\n"
            "AAAAGGCCGAAACCCC\n"
            "1234445690123456\n"
            ">TEST.mut.1 small fasta for testing\n"
            "AAAAGGCTTTCAAAACCCC\n"
            "1234445555690123456\n"))

    def test_transform_3(self):
        """overlapping mutations in somatic mode"""
        refseq = self.fa0

        grmseq = DipSeq(
            refseq.seqid + '.grm',
            refseq.description,
            size=refseq.seqA.shape[0] * 2,
            fold=refseq.fold)
        grmseq.transform(refseq, self.vcfgrm)
        out = StringIO()
        grmseq.write(out)
        self.assertEqual(out.getvalue(), (
            ">TEST.grm.0 small fasta for testing\n"
            "AAAAGGCCCCAAAACCCC\n"
            "123444567890123456\n"
            ">TEST.grm.1 small fasta for testing\n"
            "AAAAGGCCAAAACCCC\n"
            "1234447890123456\n"))

        somseq = DipSeq(
            refseq.seqid + '.som',
            refseq.description,
            size=refseq.seqA.shape[0] * 2,
            fold=refseq.fold)
        with captured_output() as (out, err):
            expected_msg = MSG_SKIP_MUT % {'allele': 1, 'POS': 5}
            somseq.transform(grmseq, self.vcfsom)
            self.assertEqual(err.getvalue(), expected_msg)
        out = StringIO()
        somseq.write(out)
        # somatic insertion at 4 and deletion at 5 both applied to allele 0
        # somatic deletion at 5 isn't applied to allele 1
        self.assertEqual(out.getvalue(), (
            ">TEST.som.0 small fasta for testing\n"
            "AAAATTGGCAAAACCCC\n"
            "12344444590123456\n"
            ">TEST.som.1 small fasta for testing\n"
            "AAAATTGGCCAAAACCCC\n"
            "123444447890123456\n"))

    def test_transform_4(self):
        """disallow mutations overlapping deletions in same vcf"""
        refseq = self.fa0
        mutseq = DipSeq(
            refseq.seqid + '.mut',
            refseq.description,
            size=refseq.seqA.shape[0] * 2,
            fold=refseq.fold)
        expected_msg = EXCEPT_MUT % {'POS': 7, 'OLDPOS': 5}
        with self.assertRaisesRegex(Exception, expected_msg):
            mutseq.transform(refseq, self.vcf4)


class TestTestUtils(unittest.TestCase):
    """Test the testing framework's methods"""

    @classmethod
    def setUpClass(cls):
        """Load some test resources"""
        # Notify aggressively so we know where we are if setup fails
        sys.stdout.write("\nTestTestUtils.setUpClass\n")
        sys.stdout.flush()
        cls.fq1 = path_join(dirname(__file__), 'resources/test.30x.1.fq')
        cls.fq2 = path_join(dirname(__file__), 'resources/test.30x.2.fq')
        cls.fq3 = path_join(dirname(__file__), 'resources/test.simple.1.fq')

    def test_Fastq(self):
        """Test the Fastq class"""
        fq1 = Fastq(self.fq1)
        fq3 = Fastq(self.fq3)
        self.assertEqual(len(fq1.records), 160)
        self.assertEqual(fq1.read_length, 150)
        self.assertEqual(fq1.records[-1]['id'], "@TEST.1_965_1528_e0_e2_9f/1")
        self.assertEqual(fq3.basecounts(), {'A': 128, 'C': 120, 'G': 98, 'T': 104})

    def test_forwardization_and_coverage(self):
        """Confirm that coverages calculated two ways agree"""
        fq1 = Fastq(self.fq1)
        fq2 = Fastq(self.fq2)
        fwd1 = Fastq.forwardize(fq1)
        fwd2 = Fastq.forwardize(fq2)
        low = min(fq1.minpos, fq2.minpos)
        high = max(fq1.maxpos, fq2.maxpos)
        for pos in range(low, high + 1):
            cov1 = set(r['id'] for r in fq1.coverage(pos) + fq2.coverage(pos))
            cov2 = set(r['id'] for r in fwd1.coverage(pos) + fwd2.coverage(pos))
            self.assertEqual(cov1, cov2)


class TestQasim(unittest.TestCase):
    """Test module-level functions"""

    @classmethod
    def setUpClass(cls):
        """Load some test resources"""
        # Notify aggressively so we know where we are if setup fails
        sys.stdout.write("\nTestQasim.setUpClass\n")
        sys.stdout.flush()
        cls.fa0 = path_join(dirname(__file__), 'resources/test.fa.0')
        cls.fa1 = path_join(dirname(__file__), 'resources/test.fa.1')
        cls.fa2 = path_join(dirname(__file__), 'resources/test.fa.2')
        cls.faA = path_join(dirname(__file__), 'resources/test.fa.A')
        cls.faC = path_join(dirname(__file__), 'resources/test.fa.C')
        cls.faG = path_join(dirname(__file__), 'resources/test.fa.G')
        cls.faT = path_join(dirname(__file__), 'resources/test.fa.T')
        cls.vcfgrm = path_join(dirname(__file__), 'resources/germline.vcf')
        cls.vcfsom = path_join(dirname(__file__), 'resources/somatic.vcf')
        cls.vcfindel = path_join(dirname(__file__), 'resources/indel.vcf.1')
        cls.vcfempty = path_join(dirname(__file__), 'resources/empty.vcf')
        cls.qpxml_R1 = path_join(dirname(__file__), 'resources/R1.qp.xml')
        cls.qpxml_R2 = path_join(dirname(__file__), 'resources/R2.qp.xml')
        cls.fq1 = path_join(dirname(__file__), 'resources/test.60x.1.fq')
        cls.fq2 = path_join(dirname(__file__), 'resources/test.60x.2.fq')

    def test_read_fasta(self):
        """Test read_fasta()"""
        seq = next(qasim.read_fasta(self.fa0))
        with StringIO() as out:
            seq.write(out)
            self.assertEqual(out.getvalue(), (
                ">TEST.0 small fasta for testing\n"
                "AAAACCCCAAAACCCC\n"
                "1234567890123456\n"
                ">TEST.1 small fasta for testing\n"
                "AAAACCCCAAAACCCC\n"
                "1234567890123456\n"))

    def test_randqual(self):
        """random choice from cumulative frequency distribution"""
        # dist = { cycle: (qual, cumulative_count) }
        dist = {
            # All 8 quals equally likely
            1: [(2, 1000), (8, 2000), (12, 3000), (22, 4000), (27, 5000),
                (32, 6000), (37, 7000), (41, 8000)],
            # high-end bias: 41 and 37 equally likely, others zero
            2: [(2, 0), (8, 0), (12, 0), (22, 0), (27, 0), (32, 0), (37, 4000),
                (41, 8000)],
            # very high-end bias: 41 is 7x more likely than 37, others zero
            3: [(2, 0), (8, 0), (12, 0), (22, 0), (27, 0), (32, 0), (37, 1000),
                (41, 8000)],
            # low-end bias: 2 is 7x more likely than 8, others zero
            4: [(2, 7000), (8, 8000), (12, 8000), (22, 8000), (27, 8000),
                (32, 8000), (37, 8000), (41, 8000)]
        }

        def avg(counts):
            """counts = [(qual, cumulative_count), ...]"""
            return sum(
                counts[i][0] *
                (counts[i][1] - (0 if i == 0 else counts[i - 1][1]))
                for i in range(len(counts))) / float(counts[-1][1])

        mu_1 = avg(dist[1])  # = 22.625
        mu_2 = avg(dist[2])  # = 39
        mu_3 = avg(dist[3])  # = 40.5
        mu_4 = avg(dist[4])  # = 2.75
        N = 100000
        mean_1 = sum(qasim._t_randqual(dist, 1) for i in range(N)) / float(N)
        mean_2 = sum(qasim._t_randqual(dist, 2) for i in range(N)) / float(N)
        mean_3 = sum(qasim._t_randqual(dist, 3) for i in range(N)) / float(N)
        mean_4 = sum(qasim._t_randqual(dist, 4) for i in range(N)) / float(N)
        self.assertAlmostEqual(mean_1 / mu_1, 1.0, delta=0.01)
        self.assertAlmostEqual(mean_2 / mu_2, 1.0, delta=0.01)
        self.assertAlmostEqual(mean_3 / mu_3, 1.0, delta=0.01)
        self.assertAlmostEqual(mean_4 / mu_4, 1.0, delta=0.01)

    def test_gen_quals(self):
        """check that P values match Q scores, and sample is representative"""
        read_length = 150
        num_quals = 10000
        qvals = np.ndarray((num_quals, read_length), dtype='u1', order="C")
        pvals = np.ndarray((num_quals, read_length), order="C")
        qasim.gen_quals(self.qpxml_R1, read_length, num_quals, qvals, pvals)

        for sample in range(num_quals):
            for q, p in zip(qvals[sample], pvals[sample]):
                # Phred definition
                self.assertEqual(p, 10 ** (q / -10))

        doc = ET.parse(self.qpxml_R1)
        Q = doc.getroot().find('.//QUAL')
        for cyclenum in range(1, read_length + 1):
            cycle = Q.find(".//Cycle[@value='%s']" % cyclenum)
            weights = sum(int(t.get('value')) * int(t.get('count'))
                          for t in cycle.findall('TallyItem'))
            counts = sum(int(t.get('count'))
                         for t in cycle.findall('TallyItem'))
            mu_qual = weights / float(counts)               # population mean
            samples = qvals[:, cyclenum - 1]
            mean_qual = sum(samples) / float(len(samples))  # sample mean
            self.assertAlmostEqual(mean_qual / mu_qual, 1.0, delta=0.01)

    def test_integration_0(self):
        """check for any & all changes to output"""
        out1 = path_join(dirname(__file__), "test_integration_0.1.fq")
        out2 = path_join(dirname(__file__), "test_integration_0.2.fq")
        read_length = 150
        contamination = 0.3
        args = [
            "--seed", "12345678",
            "--sample-name", "c9a6be94-bdb7-4c0d-a89d-4addbf76e486",
            "--vcf-input", self.vcfgrm,
            "--somatic-mode",
            "--sample-name2", "d44d739c-0143-4350-bba5-72dd068e05fd",
            "--contamination", str(contamination),
            "--vcf-input2", self.vcfsom,
            "--num-pairs", "320",
            "--quals-from", self.qpxml_R1, self.qpxml_R2,
            "--length1", str(read_length),
            "--length2", str(read_length),
            self.fa1,
            out1,
            out2]
        qasim.workflow(qasim.get_args(args))
        # There's no complicated logic here, just run a deterministic workflow
        # and compare the output to what it was when we wrote the tests.
        with open(out1) as test, open(self.fq1) as original:
            test_content = test.readlines()
            original_content = original.readlines()
            self.assertEqual(test_content, original_content)
        with open(out2) as test, open(self.fq2) as original:
            test_content = test.readlines()
            original_content = original.readlines()
            self.assertEqual(test_content, original_content)

    def test_integration_1(self):
        """check reads are generated correctly over indels"""
        out1 = path_join(dirname(__file__), "test_integration_1.1.fq")
        out2 = path_join(dirname(__file__), "test_integration_1.2.fq")
        read_length = 150
        # generate reads with no sequencing errors to make comparison back
        # to reference easy:
        args = [
            "--seed", "12345678",
            "--sample-name", "c9a6be94-bdb7-4c0d-a89d-4addbf76e486",
            "--vcf-input", self.vcfindel,
            "--num-pairs", "320",
            "--error-rate", "0",
            "--length1", str(read_length),
            "--length2", str(read_length),
            self.fa2,
            out1,
            out2]
        qasim.workflow(qasim.get_args(args))
        ref = next(qasim.read_fasta(self.fa2))
        # pad with ' ' to make 1-based reference sequence as a string
        ref_seq = ' ' + ''.join(base(b) for b in ref.seqA)
        self.assertEqual(ref_seq[1:33], "AAAAAAAACCCCCCCCGGGGGGGGTTTTTTTT")
        fq1 = Fastq(out1)
        fq2 = Fastq(out2)
        # The VCF specifies a homozygous insertion A>ACG at POS=401:
        pos = 401
        ins_reads = fq1.coverage(pos) + fq2.coverage(pos)

        # Check the forward reads over the insertion
        fwd_reads = [r for r in ins_reads if
                     r['read'] == 1 and r['frag_start'] < r['frag_end'] or
                     r['read'] == 2 and r['frag_start'] > r['frag_end']]
        for r in fwd_reads:
            read_start = min(r['frag_start'], r['frag_end'])
            for i, b in enumerate(r['seq']):
                b_pos = read_start + i
                if b_pos <= pos:
                    self.assertEqual(b, ref_seq[b_pos])
                elif b_pos == pos + 1:
                    self.assertEqual(b, 'C')
                elif b_pos == pos + 2:
                    self.assertEqual(b, 'G')
                else:
                    self.assertEqual(b, ref_seq[b_pos - 2])

        # Check the reverse reads over the insertion
        rev_reads = [r for r in ins_reads if
                     r['read'] == 1 and r['frag_start'] > r['frag_end'] or
                     r['read'] == 2 and r['frag_start'] < r['frag_end']]
        for r in rev_reads:
            read_start = max(r['frag_start'], r['frag_end'])
            if read_start == pos :
                # skip reverse reads whose start coord (far end) is /exactly/
                # pos since the indel isn't actually contained in these reads
                continue
            for i, b in enumerate(r['seq']):
                # b_pos is decreasing as we read backwards
                b_pos = read_start - i
                b = Fastq.complement[b]
                if b_pos > pos:
                    self.assertEqual(b, ref_seq[b_pos])
                elif b_pos == pos:
                    self.assertEqual(b, 'G')
                elif b_pos == pos - 1:
                    self.assertEqual(b, 'C')
                else:
                    self.assertEqual(b, ref_seq[b_pos + 2])

        # The VCF specifies a homozygous deletion AAA>A at POS=1201.
        pos = 1201
        del_reads = fq1.coverage(pos) + fq2.coverage(pos)

        # Check the forward reads over the deletion
        fwd_reads = [r for r in del_reads if
                     r['read'] == 1 and r['frag_start'] < r['frag_end'] or
                     r['read'] == 2 and r['frag_start'] > r['frag_end']]
        for r in fwd_reads:
            read_start = min(r['frag_start'], r['frag_end'])
            for i, b in enumerate(r['seq']):
                b_pos = read_start + i
                if b_pos <= pos:
                    self.assertEqual(b, ref_seq[b_pos])
                else:
                    self.assertEqual(b, ref_seq[b_pos + 2])

        # Check the reverse reads over the deletion
        rev_reads = [r for r in del_reads if
                     r['read'] == 1 and r['frag_start'] > r['frag_end'] or
                     r['read'] == 2 and r['frag_start'] < r['frag_end']]
        for r in rev_reads:
            read_start = max(r['frag_start'], r['frag_end'])
            if read_start == pos :
                # skip reverse reads whose start coord (far end) is /exactly/
                # pos since the indel isn't actually contained in these reads
                continue
            for i, b in enumerate(r['seq']):
                # b_pos is decreasing as we read backwards
                b_pos = read_start - i
                b = Fastq.complement[b]
                if b_pos >= pos + 2:
                    self.assertEqual(b, ref_seq[b_pos])
                else:
                    self.assertEqual(b, ref_seq[b_pos - 2])

    def test_integration_2(self):
        """germline mode with mutations specified by input VCF"""
        # We take advantage of the fact that we know the true location
        # of the generated reads on the reference (from the coord1_coord2
        # embedded in read ids) to check SNP genotypes at positions without
        # having to align the reads first. This wouldn't be straightforward
        # for indels because the insertion/deletion shifts the coordinates.
        out1 = path_join(dirname(__file__), "test_integration_2.1.fq")
        out2 = path_join(dirname(__file__), "test_integration_2.2.fq")
        read_length = 150
        args = [
            "--seed", "12345678",
            "--sample-name", "c9a6be94-bdb7-4c0d-a89d-4addbf76e486",
            "--vcf-input", self.vcfgrm,
            "--num-pairs", "640",
            "--quals-from", self.qpxml_R1, self.qpxml_R2,
            "--length1", str(read_length),
            "--length2", str(read_length),
            self.fa1,
            out1,
            out2]
        qasim.workflow(qasim.get_args(args))
        # work with "forwardized" reads: it's more convenient to only deal
        # with variants relative to the reference strand.
        fq1 = Fastq.forwardize(Fastq(out1))
        fq2 = Fastq.forwardize(Fastq(out2))
        # In the assertions below we're quite lenient to account for both
        # sequencing errors (introducing non-REF/ALT bases) and in the case
        # of the first two het positions, imbalanced read coverage of the
        # A & B alleles. The variants here are specified in `self.vcfgrm`
        delta = 0.1
        # A>C 0|1 SNP at position 81
        pos = 81
        covering_reads = fq1.coverage(pos) + fq2.coverage(pos)
        pos_bases = [r['seq'][pos - r['read_start']] for r in covering_reads]
        frac_A = pos_bases.count('A')/float(len(pos_bases))
        frac_C = pos_bases.count('C')/float(len(pos_bases))
        self.assertAlmostEqual(frac_A, 0.5, delta=delta)
        self.assertAlmostEqual(frac_C, 0.5, delta=delta)
        # A>C 1|0 SNP at position 161
        pos = 161
        covering_reads = fq1.coverage(pos) + fq2.coverage(pos)
        pos_bases = [r['seq'][pos - r['read_start']] for r in covering_reads]
        frac_A = pos_bases.count('A')/float(len(pos_bases))
        frac_C = pos_bases.count('C')/float(len(pos_bases))
        self.assertAlmostEqual(frac_A, 0.5, delta=delta)
        self.assertAlmostEqual(frac_C, 0.5, delta=delta)
        # A>C 1|1 SNP at position 241
        pos = 241
        covering_reads = fq1.coverage(pos) + fq2.coverage(pos)
        pos_bases = [r['seq'][pos - r['read_start']] for r in covering_reads]
        frac_A = pos_bases.count('A')/float(len(pos_bases))
        frac_C = pos_bases.count('C')/float(len(pos_bases))
        self.assertAlmostEqual(frac_A, 0.0, delta=delta)
        self.assertAlmostEqual(frac_C, 1.0, delta=delta)

    def test_integration_3(self):
        """somatic mode with mutations specified by input VCFs"""
        out1 = path_join(dirname(__file__), "test_integration_3.1.fq")
        out2 = path_join(dirname(__file__), "test_integration_3.2.fq")
        read_length = 150
        contamination = 0.3
        args = [
            "--seed", "12345678",
            "--sample-name", "c9a6be94-bdb7-4c0d-a89d-4addbf76e486",
            "--vcf-input", self.vcfgrm,
            "--somatic-mode",
            "--sample-name2", "d44d739c-0143-4350-bba5-72dd068e05fd",
            "--contamination", str(contamination),
            "--vcf-input2", self.vcfsom,
            "--num-pairs", "640",
            "--quals-from", self.qpxml_R1, self.qpxml_R2,
            "--length1", str(read_length),
            "--length2", str(read_length),
            self.fa1,
            out1,
            out2]
        qasim.workflow(qasim.get_args(args))
        # see comments in test_integration_2
        fq1 = Fastq.forwardize(Fastq(out1))
        fq2 = Fastq.forwardize(Fastq(out2))
        delta = 0.1
        # Verify that a germline variant is still present in the somatic reads
        # A>C 1|0 SNP at position 161
        pos = 161
        covering_reads = fq1.coverage(pos) + fq2.coverage(pos)
        pos_bases = [r['seq'][pos - r['read_start']] for r in covering_reads]
        frac_A = pos_bases.count('A')/float(len(pos_bases))
        frac_C = pos_bases.count('C')/float(len(pos_bases))
        self.assertAlmostEqual(frac_A, 0.5, delta=delta)
        self.assertAlmostEqual(frac_C, 0.5, delta=delta)
        # A>ACG 0|1 insertion at position 881
        pos = 881
        # We look at only "original" forward reads because in the case of the
        # first insertion specified by the somatic vcf their start coordinate
        # is unshifted from the reference, and we can perform naive position
        # arithmetic to obtain the values of the bases at pos, pos+1 & pos+2.
        #
        # Contrast this to reverse reads where the end coordinate we get
        # from the read id is /after/ the insertion and all read positions
        # relative to it are shifted by len(insert_size).
        #
        # Considering only fwd reads makes this an imperfect test of reads
        # generated over indels but a better one will require proper
        # alignment to the reference.
        fwd_covering_reads = [
            r for r in fq1.coverage(pos) + fq2.coverage(pos)
            if r['read'] == 1 and r['frag_start'] < r['frag_end'] or
            r['read'] == 2 and r['frag_start'] > r['frag_end']]
        pos1_bases = [r['seq'][pos - r['read_start']]
                      for r in fwd_covering_reads]
        pos2_bases = [r['seq'][pos + 1 - r['read_start']]
                      for r in fwd_covering_reads
                      if pos + 1 - r['read_start'] < read_length]
        pos3_bases = [r['seq'][pos + 2 - r['read_start']]
                      for r in fwd_covering_reads
                      if pos + 2 - r['read_start'] < read_length]
        frac_A1 = pos1_bases.count('A')/float(len(pos1_bases))
        self.assertAlmostEqual(frac_A1, 1.0, delta=delta)
        frac_A2 = pos2_bases.count('A')/float(len(pos2_bases))
        self.assertAlmostEqual(frac_A2, 0.5 * (1 + contamination), delta=delta)
        frac_C2 = pos2_bases.count('C')/float(len(pos2_bases))
        self.assertAlmostEqual(frac_C2, 0.5 * (1 - contamination), delta=delta)
        frac_A3 = pos3_bases.count('A')/float(len(pos3_bases))
        self.assertAlmostEqual(frac_A3, 0.5 * (1 + contamination), delta=delta)
        frac_G3 = pos3_bases.count('G')/float(len(pos3_bases))
        self.assertAlmostEqual(frac_G3, 0.5 * (1 - contamination), delta=delta)

    def test_integration_4(self):
        """check sample degradation conversion options"""
        out1 = path_join(dirname(__file__), "test_integration_4.1.fq")
        out2 = path_join(dirname(__file__), "test_integration_4.2.fq")
        bases = ['A', 'C', 'T', 'G']
        # these input fastas consist of a single base, repeated.
        fastas = [self.faA, self.faC, self.faT, self.faG]
        # conversion rate (probability)
        rate = 0.1
        for idx, (from_base, fasta) in enumerate(zip(bases, fastas)):
            from_comp = bases[idx - 2]
            for to_base in bases:
                if to_base != from_base:
                    conv = "--{}{}".format(from_base, to_base)
                    with self.subTest(msg="{} {}".format(conv, rate)):
                        # generate reads with no mutations or sequencing errors
                        args = [
                            "--seed", "12345",
                            "--vcf-input", self.vcfempty,
                            "--num-pairs", "5000",
                            "--error-rate", "0",
                            conv, str(rate),
                            fasta,
                            out1,
                            out2]
                        qasim.workflow(qasim.get_args(args))
                        counts1 = Fastq(out1).basecounts()
                        counts2 = Fastq(out2).basecounts()
                        counts = {b: counts1.get(b, 0) + counts2.get(b, 0)
                                for b in bases}
                        rate_calc = counts[to_base]/(counts[to_base] +
                                                     counts[from_base])
                        # we expect half the reads to be the reverse complement
                        # base: if that is the same as 'to_base' then
                        # conversion rate calculation is slightly different:
                        if to_base == from_comp:
                            rate_calc = 2 * rate_calc - 1
                        self.assertAlmostEqual(rate/rate_calc, 1.0, delta=0.01)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
import sys
import os, os.path
from setuptools import setup, find_packages
from distutils.extension import Extension
from distutils.dep_util import newer_group

try:
    from numpy import get_include as np_get_include
except ImportError:
    print("This extension requires numpy.")
    sys.exit(1)
try:
    from Cython.Distutils import build_ext
except ImportError:
    print("Cython not present, compiling from distributed .c")
    HAVECYTHON = False
else:
    print("Cython present, building from .pyx")
    HAVECYTHON = True

sources_cython = [os.path.join('cython','qasim.pyx'),
                  os.path.join('cython','genreads.c'),]
sources_nocyth = [os.path.join('cython','qasim.c'),
                  os.path.join('cython','genreads.c'),]


args = sys.argv[1:]
# If command is sdist, bail if no Cython available, or if qasim.c is out
# of date - distribution must have up-to-date qasim.c file. 
if args.count('sdist'):
    if not HAVECYTHON:
        print("sdist for this package requires Cython.")
        sys.exit(3)
        
    if newer_group(sources_cython, os.path.join('cython','qasim.c')):
        print("qasim.c is out of date\nRun 'setup.py build_ext' "
              "before creating distribution")
        sys.exit(4)


name = 'Qasim'
cmdclass = {}
requires = ['numpy', 'setuptools']
include_dirs = ['cython', np_get_include()]
ext_modules = []
packages = ['qasim']
package_dir = {'qasim':'qasim'}
version = '1.9.3'
author = 'Conrad Leonard'
author_email = 'conrad.leonard@hotmail.com'
platforms = ['linux']
ext_qasim = 'qasim.qasim'
license = 'MIT'
keywords = ['bioinformatics', 'simulation']
url = "https://github.com/delocalizer/qasim"
description = "Generate diploid mutations and simulate HTS reads"
scripts = ["scripts/qasim_cli.py"]


if HAVECYTHON:
    e = Extension(ext_qasim, sources_cython, include_dirs)
    e.cython_directives = {"boundscheck":False}
    cmdclass.update({'build_ext': build_ext})
    ext_modules.append(e)
else:
    e = Extension(ext_qasim, sources_nocyth, include_dirs)
    ext_modules.append(e)


setup( 
    name = name, 
    cmdclass = cmdclass,
    requires = requires,
    ext_modules = ext_modules,
    packages = packages,
    package_dir = package_dir,
    version = version,
    author = author,
    author_email = author_email,
    platforms = platforms,
    test_suite = 'tests',
    license = license,
    url = url,
    description = description,
    long_description = description,
    scripts = scripts
)
